# == Copyright: 2017, CCX Technologies

from adbus.server.object import Object
from adbus.server.method import Method
from adbus.server.method import method
from adbus.server.property import Property
from adbus.server.signal import Signal



# Copyright: 2017, CCX Technologies
#cython: language_level=3

import re
from adbus cimport sdbus_h

from typing import _GenericAlias
from typing import Any
from asyncio import get_event_loop
from asyncio import Event
from asyncio import wait_for
from asyncio import ensure_future
from errno import errorcode
from libc cimport stdint
from libc cimport stdio
from cpython.mem cimport PyMem_Malloc
from cpython.mem cimport PyMem_Free
from libc.string cimport memcpy
from libc.string cimport strncpy
from libc.string cimport strlen
from libc.string cimport memset
from cpython.ref cimport PyObject
from cpython.ref cimport Py_INCREF
from cpython.ref cimport Py_DECREF
from cpython cimport bool
from .exceptions import BusError

cdef class SdbusError(Exception):
    """SD-Bus Library Configuration Error"""

    def __init__(self, message, errno=1):
        Exception.__init__(self, message)
        self.errno = errno

include "sdbus/snake.pxi"
include "sdbus/signature.pxi"
include "sdbus/message.pxi"
include "sdbus/error.pxi"
include "sdbus/service.pxi"
include "sdbus/object.pxi"
include "sdbus/manager.pxi"
include "sdbus/method.pxi"
include "sdbus/property.pxi"
include "sdbus/signal.pxi"
include "sdbus/call.pxi"
include "sdbus/listen.pxi"



# Copyright: 2017, CCX Technologies
#cython: language_level=3

cdef bytes signature_variant = int(sdbus_h.SD_BUS_TYPE_VARIANT).to_bytes(1, 'big')
cdef bytes signature_boolean = int(sdbus_h.SD_BUS_TYPE_BOOLEAN).to_bytes(1, 'big')
cdef bytes signature_byte = int(sdbus_h.SD_BUS_TYPE_BYTE).to_bytes(1, 'big')
cdef bytes signature_int = int(sdbus_h.SD_BUS_TYPE_INT32).to_bytes(1, 'big')
cdef bytes signature_float = int(sdbus_h.SD_BUS_TYPE_DOUBLE).to_bytes(1, 'big')
cdef bytes signature_string = int(sdbus_h.SD_BUS_TYPE_STRING).to_bytes(1, 'big')
cdef bytes signature_array = int(sdbus_h.SD_BUS_TYPE_ARRAY).to_bytes(1, 'big')
cdef bytes signature_dict_begin = int(sdbus_h.SD_BUS_TYPE_DICT_ENTRY_BEGIN).to_bytes(1, 'big')
cdef bytes signature_dict_end = int(sdbus_h.SD_BUS_TYPE_DICT_ENTRY_END).to_bytes(1, 'big')
cdef bytes signature_struct_begin = int(sdbus_h.SD_BUS_TYPE_STRUCT_BEGIN).to_bytes(1, 'big')
cdef bytes signature_struct_end = int(sdbus_h.SD_BUS_TYPE_STRUCT_END).to_bytes(1, 'big')
cdef bytes signature_invalid = b'E'

cdef bytes _object_signature_basic(object obj):
    if (obj == bool) or isinstance(obj, bool):
        return signature_boolean
    elif (obj == int) or isinstance(obj, int):
        return signature_int
    elif (obj == float) or isinstance(obj, float):
        return signature_float
    elif (obj == str) or isinstance(obj, str):
        return signature_string
    elif (obj == bytes) or isinstance(obj, bytes):
        return signature_string
    elif obj == Any:
        return signature_variant
    return signature_invalid

cdef bytes _object_signature(object obj):
    cdef bytes signature = b''

    if obj is None:
        return signature

    if hasattr(obj, 'dbus_signature'):
        return obj.dbus_signature.encode('utf-8')

    elif isinstance(obj, dict):
        signature += signature_array
        signature += signature_dict_begin
        signature += _object_signature_basic(next(iter(obj.keys())))
        signature += _object_signature(next(iter(obj.values())))
        signature += signature_dict_end

    elif isinstance(obj, list):
        try:
            if all(isinstance(v, type(obj[0])) for v in obj):
                # if all the same type it's an array
                signature += signature_array
                signature += _object_signature(obj[0])
            else:
                # otherwise it's a struct
                signature += signature_struct_begin
                for v in obj:
                    signature += _object_signature(v)
                signature += signature_struct_end
        except:
            signature += signature_string

    elif isinstance(obj, tuple):
        signature += signature_struct_begin
        for v in obj:
            signature += _object_signature(v)
        signature += signature_struct_end

    elif isinstance(obj, _GenericAlias) and (obj.__origin__ == dict):
        signature += signature_array
        signature += signature_dict_begin
        signature += _object_signature_basic(obj.__args__[0])
        signature += _object_signature(obj.__args__[1])
        signature += signature_dict_end

    elif isinstance(obj, _GenericAlias) and (obj.__origin__ == list):
        signature += signature_array
        signature += _object_signature(obj.__args__[0])

    elif isinstance(obj, _GenericAlias) and (obj.__origin__ == tuple):
        signature += signature_struct_begin
        for v in obj.__args__:
            signature += _object_signature(v)
        signature += signature_struct_end

    else:
        signature += _object_signature_basic(obj)

    return signature

def variant_signature():
    return signature_variant.decode()

cdef bytes _dbus_signature(obj):
    cdef bytes signature = _object_signature(obj)

    if signature_invalid in signature:
        raise TypeError(f"No D-Bus type equivalent for {obj}")

    return signature

def dbus_signature(obj):
    """Calculates a D-Bus Signature from a Python object or type.

    Args:
        obj (object or type): Python object or type,
            If the object has a dbus_signature attribute it will
            be used, otherwise the object or type will be parsed
            to calculate the D-Bus Signature.
            Supports bool, int, str, float, bytes, and from the
            typing library, List, Dict, and Tuple.

    Returns:
        A string representing the D-Bus Signature.

    Raises:
        TypeError: If no D-Bus Equivalent for objects type.
    """
    signature = _dbus_signature(obj)

    try:
        return signature.decode()
    except UnicodeDecodeError:
        # There is a memory management bug that rarely results in a Unicode Decode
        # error, this was added to get a little more info when it happens, but it
        # hasn't happened since this was added...
        import syslog
        syslog.syslog(f"==> {obj} === {signature}")
        raise

cdef object _object_cast_basic(bytes signature, object obj):
    if signature[0] == sdbus_h.SD_BUS_TYPE_BOOLEAN :
        return bool(obj)
    elif signature[0] == sdbus_h.SD_BUS_TYPE_INT32:
        return int(obj)
    elif signature[0] == sdbus_h.SD_BUS_TYPE_DOUBLE:
        return float(obj)
    elif signature[0] == sdbus_h.SD_BUS_TYPE_STRING:
        if type(obj) == bytes:
            return obj.decode('utf-8', errors='ignore')
        return str(obj)
    return obj

cdef object _object_cast(bytes signature, object obj):

    if hasattr(obj, 'dbus_value'):
        obj = obj.dbus_value

    if signature[0] == sdbus_h.SD_BUS_TYPE_ARRAY:
        if signature[1] == sdbus_h.SD_BUS_TYPE_DICT_ENTRY_BEGIN:
            return {_object_cast_basic(signature[2:], k): _object_cast(signature[3:], v)
                    for k,v in obj.items()}
        else:
            return [_object_cast(signature[1:], v) for v in obj]
    elif signature[0] == sdbus_h.SD_BUS_TYPE_STRUCT_BEGIN:
        return tuple([_object_cast(signature[2:], v) for v in obj])

    else:
        return _object_cast_basic(signature, obj)

def dbus_cast(signature, obj):
    """Casts an object into the type defined in a D-Bus signature.

    Args:
        signature (str): D-Bus type signature
        obj (object or type): Python object or type

    Returns:
        A new type, cast from the D-Bus Signature.

    """
    return _object_cast(signature.encode(), obj)



# == Copyright: 2017, CCX Technologies
#cython: language_level=3

cdef union _value:
    stdint.uint8_t c_byte
    stdint.uint16_t c_uint16
    stdint.int16_t c_int16
    stdint.uint32_t c_uint32
    stdint.int32_t c_int32
    stdint.uint64_t c_uint64
    stdint.int64_t c_int64
    double c_double
    bint c_bool
    const char* c_str

cdef class Message:
    cdef sdbus_h.sd_bus_message *message

    def __cinit__(self):
        self.message = NULL

    def __dealloc__(self):
        self.message = sdbus_h.sd_bus_message_unref(self.message)

    cdef import_sd_bus_message(self, sdbus_h.sd_bus_message *message):
        cdef errno
        cdef const sdbus_h.sd_bus_error *err

        err = sdbus_h.sd_bus_message_get_error(message)
        if err:
            errno = sdbus_h.sd_bus_message_get_errno(message)
            raise SdbusError(f"{err.name.decode()}: {err.message.decode()}", -errno)

        self.message = sdbus_h.sd_bus_message_unref(self.message)
        self.message = sdbus_h.sd_bus_message_ref(message)

    cdef new_method_call(self, Service service, char *destination,
            char *path, char *interface, const char *member):
        cdef int ret

        self.message = sdbus_h.sd_bus_message_unref(self.message)
        ret = sdbus_h.sd_bus_message_new_method_call(service.bus,
            &self.message, destination, path, interface, member)
        if ret < 0:
            raise SdbusError(
                f"Failed to create new method call: {errorcode[-ret]}", -ret)

    cdef new_method_return(self):
        cdef int ret
        cdef sdbus_h.sd_bus_message *call = self.message
        ret = sdbus_h.sd_bus_message_new_method_return(call, &self.message)
        if ret < 0:
            raise SdbusError(
                f"Failed to create new method return: {errorcode[-ret]}", -ret)
        sdbus_h.sd_bus_message_unref(call)

    cdef new_method_error(self, sdbus_h.sd_bus_message *call,
            Exception exception):
        cdef int ret
        self.message = sdbus_h.sd_bus_message_unref(self.message)
        ret = sdbus_h.sd_bus_message_new_method_return(call, &self.message)
        if ret < 0:
            raise SdbusError(
                f"Failed to create new method error: {errorcode[-ret]}", -ret)

    cdef new_signal(self, Signal signal):
        cdef int ret
        self.message = sdbus_h.sd_bus_message_unref(self.message)
        ret = sdbus_h.sd_bus_message_new_signal(signal.bus, &self.message,
                signal.path, signal.interface, signal.name)
        if ret < 0:
            raise SdbusError(
                f"Failed to create new signal: {errorcode[-ret]}", -ret)

    # ------------

    cdef _element_length(self, const char *signature):
        cdef unsigned int i = 0
        cdef unsigned int scnt = 0
        cdef unsigned int dcnt = 0

        while signature[i] != 0:
            s = signature[i]
            i += 1

            if s == sdbus_h.SD_BUS_TYPE_ARRAY:
                continue
            elif s == sdbus_h.SD_BUS_TYPE_STRUCT_BEGIN:
                scnt += 1
            elif s == sdbus_h.SD_BUS_TYPE_STRUCT_END:
                scnt -= 1
            elif s == sdbus_h.SD_BUS_TYPE_DICT_ENTRY_BEGIN:
                dcnt += 1
            elif s == sdbus_h.SD_BUS_TYPE_DICT_ENTRY_END:
                dcnt -= 1

            if not dcnt and not scnt:
                break

        return i

    # ------------

    cdef _read_basic(self, char sig, void *value):
        cdef int ret
        ret = sdbus_h.sd_bus_message_read_basic(self.message, sig, value)
        if ret < 0:
            raise SdbusError(
                    f"Failed to read value with signature {chr(sig)}: {errorcode[-ret]}",
                    -ret)

    cdef _read_array(self, const char *signature, unsigned int *index):
        cdef unsigned int elength = self._element_length(&signature[index[0]])
        cdef bytes bsignature = signature
        cdef bytes psignature = bsignature[index[0]:elength+index[0]] + bytes(1)
        cdef char *esignature = psignature
        cdef list values = []
        cdef list value

        index[0] += elength

        if sdbus_h.sd_bus_message_enter_container(self.message,
                sdbus_h.SD_BUS_TYPE_ARRAY, esignature) < 0:
            raise SdbusError(f"Failed to enter array {esignature}")

        while not sdbus_h.sd_bus_message_at_end(self.message,0):
            value = self.read(esignature)
            if len(value) == 1:
                values.append(value[0])
            else:
                values.append(value)

        if sdbus_h.sd_bus_message_exit_container(self.message) < 0:
            raise SdbusError(f"Failed to exit array {esignature}")

        # A dictionary is always an array with two elements (based on D-Bus
        # definition) so if we're a dictionary convert it.
        if esignature[0] == sdbus_h.SD_BUS_TYPE_DICT_ENTRY_BEGIN:
            return {v[0]: v[1] for v in values}
        else:
            return values

    cdef _read_variant(self):
        cdef const char *esignature
        cdef list value

        if sdbus_h.sd_bus_message_enter_container(self.message,
                sdbus_h.SD_BUS_TYPE_VARIANT, NULL) < 0:
            raise SdbusError("Failed to enter variant")

        esignature = sdbus_h.sd_bus_message_get_signature(self.message, 0)
        value = self.read(esignature)

        if sdbus_h.sd_bus_message_exit_container(self.message) < 0:
            raise SdbusError(f"Failed to exit variant {esignature}")

        # variant only has one type
        return value[0]

    cdef tuple _read_struct(self, const char *signature, unsigned int *index):
        cdef unsigned int elength = self._element_length(&signature[index[0]-1])-1
        cdef bytes bsignature = signature
        cdef bytes psignature = bsignature[index[0]:elength+index[0]-1] + bytes(1)
        cdef char *esignature = psignature
        cdef list value

        index[0] += elength + 1

        if sdbus_h.sd_bus_message_enter_container(self.message,
                sdbus_h.SD_BUS_TYPE_STRUCT, esignature) < 0:
            raise SdbusError(f"Failed to enter structure {esignature}")

        value = self.read(esignature)

        if sdbus_h.sd_bus_message_exit_container(self.message) < 0:
            raise SdbusError(f"Failed to exit structure {esignature}")

        return tuple(value)

    cdef list _read_dict(self, const char *signature, unsigned int *index):
        cdef unsigned int elength = self._element_length(&signature[index[0]-1])-1
        cdef bytes bsignature = signature
        cdef bytes psignature = bsignature[index[0]:elength+index[0]-1] + bytes(1)
        cdef char *esignature = psignature

        index[0] += elength + 1

        if sdbus_h.sd_bus_message_enter_container(self.message,
                sdbus_h.SD_BUS_TYPE_DICT_ENTRY, esignature) < 0:
            raise SdbusError(f"Failed to enter dictionary {esignature}")

        value = self.read(esignature)

        if sdbus_h.sd_bus_message_exit_container(self.message) < 0:
            raise SdbusError(f"Failed to exit dictionary {esignature}")

        return value

    cdef list read(self, const char *signature=b"ANY"):
        cdef _value v
        cdef list values = []
        cdef unsigned int i = 0
        cdef char s

        if signature[0] == b'A':
            signature = sdbus_h.sd_bus_message_get_signature(self.message, 0)

        while i < len(signature):
            s = signature[i]
            i += 1

            if s ==  sdbus_h._SD_BUS_TYPE_INVALID:
                break

            elif s == sdbus_h.SD_BUS_TYPE_ARRAY:
                values.append(self._read_array(signature, &i))

            elif s == sdbus_h.SD_BUS_TYPE_VARIANT:
                values.append(self._read_variant())

            elif s == sdbus_h.SD_BUS_TYPE_STRUCT_BEGIN:
                values.append(self._read_struct(signature, &i))

            elif s == sdbus_h.SD_BUS_TYPE_DICT_ENTRY_BEGIN:
                values.append(self._read_dict(signature, &i))

            elif s == sdbus_h.SD_BUS_TYPE_BYTE:
                self._read_basic(s, <void*>&v.c_byte)
                values.append(v.c_byte)

            elif s == sdbus_h.SD_BUS_TYPE_BOOLEAN:
                self._read_basic(s, <void*>&v.c_bool)
                values.append(v.c_bool)

            elif s == sdbus_h.SD_BUS_TYPE_UINT16:
                self._read_basic(s, <void*>&v.c_uint16)
                values.append(v.c_uint16)

            elif s == sdbus_h.SD_BUS_TYPE_INT16:
                self._read_basic(s, <void*>&v.c_int16)
                values.append(v.c_int16)

            elif s == sdbus_h.SD_BUS_TYPE_UINT32:
                self._read_basic(s, <void*>&v.c_uint32)
                values.append(v.c_uint32)

            elif s == sdbus_h.SD_BUS_TYPE_INT32:
                self._read_basic(s, <void*>&v.c_int32)
                values.append(v.c_int32)

            elif s == sdbus_h.SD_BUS_TYPE_UINT64:
                self._read_basic(s, <void*>&v.c_uint64)
                values.append(v.c_uint64)

            elif s == sdbus_h.SD_BUS_TYPE_INT64:
                self._read_basic(s, <void*>&v.c_int64)
                values.append(v.c_int64)

            elif s == sdbus_h.SD_BUS_TYPE_DOUBLE:
                self._read_basic(s, <void*>&v.c_double)
                values.append(v.c_double)

            elif s == sdbus_h.SD_BUS_TYPE_STRING:
                self._read_basic(s, <void*>&v.c_str)
                values.append(v.c_str.decode('utf-8'))

            elif s == sdbus_h.SD_BUS_TYPE_OBJECT_PATH:
                self._read_basic(s, <void*>&v.c_str)
                values.append(v.c_str.decode('utf-8'))

            elif s == sdbus_h.SD_BUS_TYPE_SIGNATURE:
                self._read_basic(s, <void*>&v.c_str)
                values.append(v.c_str.decode('utf-8'))

            elif s == sdbus_h.SD_BUS_TYPE_UNIX_FD:
                self._read_basic(s, <void*>&v.c_int32)
                values.append(v.c_int32)

            else:
                raise SdbusError(f"Unsupported signature type {signature} -> {chr(s)} for read")

        return values

    # ------------

    cdef _append_basic(self, char sig, const void *value):
        cdef int ret
        ret = sdbus_h.sd_bus_message_append_basic(self.message, sig, value)
        if ret < 0:
            raise SdbusError(
                    f"Failed to append value {chr(sig)}: {errorcode[-ret]}", -ret)

    cdef _append_array(self, const char *signature, object value,
            unsigned int *index):
        cdef unsigned int elength = self._element_length(&signature[index[0]])
        cdef bytes bsignature = signature
        cdef bytes psignature = bsignature[index[0]:elength+index[0]] + bytes(1)
        cdef char *esignature = psignature

        index[0] += elength

        if sdbus_h.sd_bus_message_open_container(self.message,
                sdbus_h.SD_BUS_TYPE_ARRAY, esignature) < 0:
            raise SdbusError(f"Failed to open array {esignature}")

        for item in value:
            self.append(esignature, item)

        if sdbus_h.sd_bus_message_close_container(self.message) < 0:
            raise SdbusError(f"Failed to close array {esignature}")

    cdef _append_variant(self, object value):
        cdef bytes signature
        cdef const char *esignature

        signature = _object_signature(value)
        esignature = signature

        if sdbus_h.sd_bus_message_open_container(self.message,
                sdbus_h.SD_BUS_TYPE_VARIANT, esignature) < 0:
            raise SdbusError(f"Failed to open variant {esignature}")

        self.append(esignature, value)

        if sdbus_h.sd_bus_message_close_container(self.message) < 0:
            raise SdbusError(f"Failed to close variant {esignature}")

    cdef _append_struct(self, const char *signature, object value,
            unsigned int *index):
        cdef unsigned int elength = self._element_length(&signature[index[0]-1])-2
        cdef bytes bsignature = signature
        cdef bytes psignature = bsignature[index[0]:elength+index[0]] + bytes(1)
        cdef char *esignature = psignature

        index[0] += elength + 2

        if sdbus_h.sd_bus_message_open_container(self.message,
                sdbus_h.SD_BUS_TYPE_STRUCT, esignature) < 0:
            raise SdbusError(f"Failed to open struct {psignature} {index[0]} {elength} {signature}")

        self._append_multiple(esignature, value)

        if sdbus_h.sd_bus_message_close_container(self.message) < 0:
            raise SdbusError(f"Failed to close struct {esignature}")

    cdef _append_dict(self, const char *signature, object value,
            unsigned int *index):
        cdef unsigned int elength = self._element_length(&signature[index[0]])
        cdef bytes bsignature = signature
        cdef bytes psignature = bsignature[index[0]:elength+index[0]] + bytes(1)
        cdef char *esignature = psignature

        index[0] += elength

        if sdbus_h.sd_bus_message_open_container(self.message,
                sdbus_h.SD_BUS_TYPE_ARRAY, esignature) < 0:
            raise SdbusError(f"Failed to open dict {esignature}")

        for item in value.items():
            self._append_dict_item(esignature, item)

        if sdbus_h.sd_bus_message_close_container(self.message) < 0:
            raise SdbusError(f"Failed to close dict {esignature}")

    cdef _append_dict_item(self, const char *signature, object value):
        cdef unsigned int elength = self._element_length(&signature[2])
        cdef bytes bsignature = signature
        cdef bytes psignature = bsignature[1:2+elength] + bytes(1)
        cdef char *esignature = psignature
        cdef char ksignature = esignature[0]
        cdef char *vsignature = &esignature[1]

        if sdbus_h.sd_bus_message_open_container(self.message,
                sdbus_h.SD_BUS_TYPE_DICT_ENTRY, esignature) < 0:
            raise SdbusError(f"Failed to open dict item {esignature}")

        self.append(&ksignature, value[0])
        self.append(vsignature, value[1])

        if sdbus_h.sd_bus_message_close_container(self.message) < 0:
            raise SdbusError(f"Failed to close dict item {esignature}")

    cdef _append_multiple(self, const char *signature, object values):
        cdef unsigned int i = 0
        cdef char s
        cdef object value

        if isinstance(values, (str, bytes)):
            values = [values]
        elif isinstance(values, tuple):
            values = list(values)

        while values:
            s = signature[i]
            i += 1

            value = values.pop(0)

            try:
                value = value.dbus_value
            except AttributeError:
                pass

            if s == sdbus_h.SD_BUS_TYPE_ARRAY:
                if signature[1] == sdbus_h.SD_BUS_TYPE_DICT_ENTRY_BEGIN:
                    self._append_dict(signature, value, &i)
                else:
                    self._append_array(signature, value, &i)

            elif s == sdbus_h.SD_BUS_TYPE_STRUCT_BEGIN:
                self._append_struct(signature, value, &i)

            else:
                self.append(&s, value)

    cdef append(self, const char *signature, object value):
        cdef _value v
        cdef unsigned int i = 1
        cdef char s
        cdef bytes v_str

        s = signature[0]

        try:
            value = value.dbus_value
        except AttributeError:
            pass

        if s == sdbus_h._SD_BUS_TYPE_INVALID:
            return

        elif s == sdbus_h.SD_BUS_TYPE_ARRAY:
            if signature[1] == sdbus_h.SD_BUS_TYPE_DICT_ENTRY_BEGIN:
                self._append_dict(signature, value, &i)
            else:
                self._append_array(signature, value, &i)

        elif s == sdbus_h.SD_BUS_TYPE_VARIANT:
            self._append_variant(value)

        elif s == sdbus_h.SD_BUS_TYPE_STRUCT_BEGIN:
            self._append_struct(signature, value, &i)

        elif s == sdbus_h.SD_BUS_TYPE_BYTE:
            v.c_byte = value
            self._append_basic(s, <void*>&v.c_byte)

        elif s == sdbus_h.SD_BUS_TYPE_BOOLEAN:
            v.c_bool = value
            self._append_basic(s, <void*>&v.c_bool)

        elif s == sdbus_h.SD_BUS_TYPE_UINT16:
            v.c_uint16 = value
            self._append_basic(s, <void*>&v.c_uint16)

        elif s == sdbus_h.SD_BUS_TYPE_INT16:
            v.c_int16 = value
            self._append_basic(s, <void*>&v.c_int16)

        elif s == sdbus_h.SD_BUS_TYPE_UINT32:
            v.c_uint32 = value
            self._append_basic(s, <void*>&v.c_uint32)

        elif s == sdbus_h.SD_BUS_TYPE_INT32:
            v.c_int32 = value
            self._append_basic(s, <void*>&v.c_int32)

        elif s == sdbus_h.SD_BUS_TYPE_UINT64:
            v.c_uint64 = value
            self._append_basic(s, <void*>&v.c_uint64)

        elif s == sdbus_h.SD_BUS_TYPE_INT64:
            v.c_int64 = value
            self._append_basic(s, <void*>&v.c_int64)

        elif s == sdbus_h.SD_BUS_TYPE_DOUBLE:
            v.c_double = value
            self._append_basic(s, <void*>&v.c_double)

        elif s == sdbus_h.SD_BUS_TYPE_STRING:
            if type(value) == str:
                v_str = value.encode('utf-8')
                v.c_str = v_str
            else:
                try:
                    v.c_str = value
                except TypeError as e:
                    e.args = (f"expected str or bytes, {type(value).__name__} found",)
                    raise
            self._append_basic(s, <void*>v.c_str)

        elif s == sdbus_h.SD_BUS_TYPE_OBJECT_PATH:
            if type(value) == str:
                v_str = value.encode('utf-8')
                v.c_str = v_str
            else:
                try:
                    v.c_str = value
                except TypeError as e:
                    e.args = (f"expected str or bytes, {type(value).__name__} found",)
                    raise
            self._append_basic(s, <void*>v.c_str)

        elif s == sdbus_h.SD_BUS_TYPE_SIGNATURE:
            if type(value) == str:
                v_str = value.encode('utf-8')
                v.c_str = v_str
            else:
                try:
                    v.c_str = value
                except TypeError as e:
                    e.args = (f"expected str or bytes, {type(value).__name__} found",)
                    raise
            self._append_basic(s, <void*>v.c_str)

        elif s == sdbus_h.SD_BUS_TYPE_UNIX_FD:
            v.c_int32 = value
            self._append_basic(s, <void*>&v.c_int32)

        else:
            raise SdbusError(f"Unsupported signature type {chr(s)} for append")


    # ------------

    cdef send(self):
        cdef int ret
        ret = sdbus_h.sd_bus_send(NULL, self.message, NULL)
        if ret < 0:
            raise SdbusError(f"Failed to send message: {errorcode[-ret]}", -ret)




# Copyright: 2017, CCX Technologies
"""D-Bus Exceptions from the adbus package"""


class BusError(Exception):
    """Error Interfacing on the D-Bus

    Args:
        message: Error message
    """
    pass



# == Copyright: 2017, CCX Technologies
#cython: language_level=3

cdef int property_get_handler(sdbus_h.sd_bus *bus,
            const char *path, const char *interface, const char *propname,
            sdbus_h.sd_bus_message *m, void *userdata, sdbus_h.sd_bus_error *err):
    cdef PyObject *property_ptr = (((<PyObject**>userdata)[0]))
    cdef Property property = <Method>property_ptr
    cdef Message message = Message()
    cdef object value
    cdef bytes err_name
    cdef bytes err_message

    message.import_sd_bus_message(m)

    try:
        value = getattr(<object>(property.instance), property.attr_name)
        message.append(property.signature, value)
    except Exception as e:
        property.loop.call_exception_handler({'message': f"{e} for {property.attr_name} <{property.signature}>", 'exception': e})

		#NOTE: sd_bus uses the System.Error prefix for general purpose errors
        err_name = b"System.Error." + \
            e.__class__.__name__.encode('utf-8')
        err_message = f"{e} for {property.attr_name} <{property.signature}>".encode('utf-8')
        sdbus_h.sd_bus_error_set(err, err_name, err_message)

        return -1

    return 1

cdef int property_set_handler(sdbus_h.sd_bus *bus,
            const char *path, const char *interface, const char *propname,
            sdbus_h.sd_bus_message *m, void *userdata, sdbus_h.sd_bus_error *err):
    cdef PyObject *property_ptr = (((<PyObject**>userdata)[0]))
    cdef Property property = <Method>property_ptr
    cdef Message message = Message()
    cdef Error error
    cdef list values
    cdef bytes err_name
    cdef bytes err_message

    message.import_sd_bus_message(m)
    values = message.read(property.signature)

    try:
        # If the old value was created with a dbus_value property then
        # need to create a new instance of that object
        old_value = getattr(<object>(property.instance), property.attr_name)
        if hasattr(old_value, 'dbus_value'):
            value = type(old_value)(values[0])
        else:
            value = values[0]

        setattr(<object>(property.instance), property.attr_name, value)
    except Exception as e:
        property.loop.call_exception_handler({'message': f"{e} for {property.attr_name}", 'exception': e})

		#NOTE: sd_bus uses the System.Error prefix for general purpose errors
        err_name = b"System.Error." + \
            e.__class__.__name__.encode('utf-8')
        err_message = f"{e} for {property.attr_name}".encode('utf-8')
        sdbus_h.sd_bus_error_set(err, err_name, err_message)

        return -1

    return 1

cdef class Property:
    cdef stdint.uint8_t type
    cdef stdint.uint64_t flags
    cdef sdbus_h.sd_bus_vtable_property x
    cdef void *userdata
    cdef object py_instance
    cdef PyObject *instance
    cdef str attr_name
    cdef bytes name
    cdef bytes signature
    cdef bool connected
    cdef object loop

    def __cinit__(self, name, py_object, attr_name, signature='', read_only=False,
            deprecated=False, hidden=False, unprivileged=False,
            emits_constant=False, emits_change=False, emits_invalidation=False):

        self.name = name.encode()
        self.py_instance = py_object
        self.attr_name = attr_name
        self.signature = signature.encode()
        self.connected = False

        if read_only:
            self.type = sdbus_h._SD_BUS_VTABLE_PROPERTY
        else:
            self.type = sdbus_h._SD_BUS_VTABLE_WRITABLE_PROPERTY

        self.flags = 0
        if deprecated:
            self.flags |= sdbus_h.SD_BUS_VTABLE_DEPRECATED

        if hidden:
            self.flags |= sdbus_h.SD_BUS_VTABLE_HIDDEN

        if unprivileged:
            self.flags |= sdbus_h.SD_BUS_VTABLE_UNPRIVILEGED

        if emits_constant:
            self.flags |= sdbus_h.SD_BUS_VTABLE_PROPERTY_CONST

        if emits_change:
            self.flags |= sdbus_h.SD_BUS_VTABLE_PROPERTY_EMITS_CHANGE

        if emits_invalidation:
            self.flags |= sdbus_h.SD_BUS_VTABLE_PROPERTY_EMITS_INVALIDATION

        self.x.member = self.name
        self.x.get = property_get_handler
        if not read_only:
            self.x.set = property_set_handler
        else:
            self.x.set = NULL
        self.x.signature = self.signature

        self.userdata = <void *>self

    cdef populate_vtable(self, sdbus_h.sd_bus_vtable *vtable):
        vtable.type = self.type
        vtable.flags = self.flags
        memcpy(&vtable.x, &self.x, sizeof(self.x))

    cdef set_object(self, Object object):
        if self.connected:
            raise SdbusError("Property already associated")
        self.connected = True

        # this decrements the counter of py_instance so if it
        # refers to the base object the object won't be held in
        # memory indefinitely (self-referenced)
        self.instance = <PyObject *>self.py_instance
        self.py_instance = None

        self.loop = object.loop

    def emit_changed(self):
        if not self.object:
            raise SdbusError("Property not associated")
        return self.object.emit_properties_changed([self.name.decode()])



# Copyright: 2017-2021, CCX Technologies
"""D-Bus Proxy"""

import xml.etree.ElementTree as etree
import typing

from .. import sdbus
from .. import exceptions
from . import call
from . import get
from . import get_all
from . import set_
from . import Listen


class _Empty_Class:
    pass


class _DbusWrapper:
    def __init__(self, signature, value):
        self.dbus_signature = signature
        self.dbus_value = value


class Signal:
    def __init__(
            self, parent, service, address, path, interface, etree, timeout_ms
    ):
        self.parent = parent
        self.service = service
        self.address = address
        self.path = path
        self.interface = interface
        self.timeout_ms = timeout_ms
        self.signature = ''
        self.listens = {}

        for x in etree.iter():
            if x.tag == 'signal':
                self.name = x.attrib['name']
            if x.tag == 'arg':
                self.signature += x.attrib['type']

    def add(self, coroutine, signature=False):
        """Add a listening coroutine to this signal.

        Args:
            coroutine (coroutine): coroutine to schedule when signal received
            signature (str): optional, signature of the signal, if False the
                types of the coroutine arguments will be used to create the
                signature and the coroutine will be called with one argument
                per signal argument, if defined the coroutine will be called
                with a list of arguments, if None the coroutine will be called
                with a list of types determined at run-time
        """
        listen = Listen(
                self.service,
                self.address,
                self.path,
                self.interface,
                self.name,
                coroutine,
                signature=signature
        )

        if (
                (listen.signature != 'ANY')
                and (self.signature != listen.signature)
        ):
            raise exceptions.BusError(
                    f"Coroutine signature {listen.signature} doesn't "
                    f"match signal signature {self.signature}."
            )

        self.listens[coroutine.__name__] = listen

    def remove(self, coroutine):
        del self.listens[coroutine.__name__]

    def __call__(self, coroutine, remove=True):
        if remove and (coroutine.__name__ in self.listens):
            self.remove(coroutine)
        else:
            self.add(coroutine)


class Property:
    emits_changed_signal = 'true'
    cached_value = None

    def __init__(
            self, parent, service, address, path, interface, etree, timeout_ms
    ):
        self.parent = parent
        self.service = service
        self.address = address
        self.path = path
        self.interface = interface
        self.timeout_ms = timeout_ms
        self.name = etree.attrib['name']
        self.signature = etree.attrib['type']
        self.trackers = {}

        if etree.attrib['access'] == 'readwrite':
            self.readonly = False
        else:
            self.readonly = True

        for x in etree:
            if x.tag == 'annotation':
                if (
                        x.attrib['name'] ==
                        'org.freedesktop.DBus.Property.EmitsChangedSignal'
                ):
                    self.emits_changed_signal = x.attrib['value'].lower()

    async def get(self):
        if (
                (self.emits_changed_signal == 'false')
                or (self.cached_value is None)
        ):
            self.cached_value = await get(
                    self.service,
                    self.address,
                    self.path,
                    self.interface,
                    self.name,
                    timeout_ms=self.timeout_ms
            )

        return self.cached_value

    async def set(self, value):
        if self.emits_changed_signal == 'const':
            raise AttributeError(f"Can't set read-only property {self.name}")
        else:
            try:
                await set_(
                        self.service,
                        self.address,
                        self.path,
                        self.interface,
                        self.name,
                        sdbus.dbus_cast(self.signature, value),
                        timeout_ms=self.timeout_ms
                )
            except sdbus.SdbusError:
                # sometimes we'll get a EINTR (like one in a million trys), we want
                # to try again if we do
                await set_(
                        self.service,
                        self.address,
                        self.path,
                        self.interface,
                        self.name,
                        sdbus.dbus_cast(self.signature, value),
                        timeout_ms=self.timeout_ms
                )

    def track(self, coroutine):
        self.trackers[coroutine.__name__] = coroutine

    def untrack(self, coroutine):
        del self.trackers[coroutine.__name__]

    async def __call__(self, value=None):
        if value is None:
            return await self.get()
        else:
            await self.set(value)
            return await self.get()


class Method:
    def __init__(
            self, parent, service, address, path, interface, etree, timeout_ms
    ):
        self.parent = parent
        self.service = service
        self.address = address
        self.path = path
        self.interface = interface
        self.timeout_ms = timeout_ms
        self.arg_signatures = []
        self.return_signature = ''

        for x in etree.iter():
            if x.tag == 'method':
                self.name = x.attrib['name']
            if x.tag == 'arg':
                if x.attrib['direction'] == 'out':
                    self.return_signature += x.attrib['type']
                elif x.attrib['direction'] == 'in':
                    self.arg_signatures.append(x.attrib['type'])

    async def __call__(self, *args):
        return await call(
                self.service,
                self.address,
                self.path,
                self.interface,
                self.name, [
                        _DbusWrapper(
                                self.arg_signatures[i],
                                sdbus.dbus_cast(self.arg_signatures[i], a)
                        ) for i, a in enumerate(args)
                ],
                self.return_signature,
                timeout_ms=self.timeout_ms
        )


class Interface:
    def __init__(
            self, parent, service, address, path, etree, camel_convert,
            timeout_ms, changed_coroutine
    ):
        self.parent = parent
        interface = etree.attrib['name']
        self.interface = interface
        self.methods = {}
        self.signals = {}
        self.properties = {}
        self.changed_coroutine = changed_coroutine
        self.camel_convert = camel_convert

        self.service = service
        self.address = address
        self.path = path
        self.interface = interface
        self.timeout_ms = timeout_ms

        def _add_snake_and_camel(d, n, v):
            d[n] = v
            if camel_convert:
                d[sdbus.camel_to_snake(n)] = v

        for m in etree.iter('method'):
            _add_snake_and_camel(
                    self.methods, m.attrib['name'],
                    Method(
                            self, service, address, path, interface, m,
                            timeout_ms
                    )
            )

        for p in etree.iter('property'):
            _add_snake_and_camel(
                    self.properties, p.attrib['name'],
                    Property(
                            self, service, address, path, interface, p,
                            timeout_ms
                    )
            )

        if self.properties:
            self.properties_changed_listen = Listen(
                    service,
                    address,
                    path,
                    "org.freedesktop.DBus.Properties",
                    "PropertiesChanged",
                    self.properties_changed,
                    args=(interface, ),
            )

        else:
            self.properties_changed_listen = None

        for s in etree.iter('signal'):
            _add_snake_and_camel(
                    self.signals, s.attrib['name'],
                    Signal(
                            self, service, address, path, interface, s,
                            timeout_ms
                    )
            )

    async def update_properties(self):
        if self.properties_changed_listen is not None:
            values = await get_all(
                    self.service, self.address, self.path, self.interface,
                    self.timeout_ms
            )
            for p, v in values.items():
                self.properties[p].cached_value = v

    async def properties_changed(
            self, interface: str, changed: typing.Dict[str, typing.Any],
            invalidated: typing.List[str]
    ):
        for p, v in changed.items():
            prop = self.properties[p]
            prop.cached_value = v
            for coroutine in prop.trackers.values():
                await coroutine(v)

        for p in invalidated:
            self.properties[p].cached_value = None

        if self.changed_coroutine:
            changed = list(changed.keys()) + invalidated
            if self.camel_convert:
                changed = [sdbus.camel_to_snake(x) for x in changed]
            await self.changed_coroutine(changed)

    def __getattr__(self, name):

        if name == 'properties':
            return self.properties

        if name == 'methods':
            return self.methods

        if name == 'signals':
            return self.signals

        try:
            return self.methods[name]
        except KeyError:
            pass

        try:
            return self.properties[name]
        except KeyError:
            pass

        try:
            return self.signals[name]
        except KeyError:
            pass

        raise AttributeError(
                f"'{self.interface}' interface has no attribute '{name}'"
        )

    def set_changed_coroutine(self, changed_coroutine):
        """Set the Interface's Changed Co-Routine."""
        self.changed_coroutine = changed_coroutine

    async def __aenter__(self):
        self._property_multi = _Empty_Class()
        return self._property_multi

    async def __aexit__(
            self, exception_type, exception_value, exception_traceback
    ):
        properties = _DbusWrapper(
                'a{sv}', {
                        sdbus.snake_to_camel(p) if self.camel_convert else p:
                        sdbus.dbus_cast(self.properties[p].signature, v)
                        for p, v in self._property_multi.__dict__.items()
                }
        )

        try:
            await self.parent["ccx.DBus.Properties"].methods["SetMulti"](
                    self.interface, properties
            )
        except KeyError:
            for p, v in properties.dbus_value.items():
                await self.properties[p].set(v)


class Proxy:
    """Creates a Python Object that maps to an Interface that already
    exists on the D-Bus.

    Args:
        service (adbus.server.Service): service to connect to
        address (str): address (name) of the application to connect to
            on the D-Bus.
        path (str): path to create on the service
            ie. /com/awesome/Settings1
        interface (str): optional, default interface to access, if None will
            use address
        changed_coroutines: optional, coroutine dictionary of changed
            coroutines, the key is the name of the interface to attach
            to. The coroutine will be called with a single argument which is a
            list of property names.
        timeout_ms (int): optional, maximum time to wait for a response in
            milli-seconds
        camel_convert (bool): optional, D-Bus method and property
            names are typically defined in Camel Case, but Python
            methods and arguments are typically defined in Snake
            Case, if this is set the cases will be automatically
            converted between the two

    """

    _interface = None
    _introspect_xml = None

    def __init__(
            self,
            service,
            address,
            path,
            interface=None,
            changed_coroutines=None,
            timeout_ms=30000,
            camel_convert=True,
    ):
        self._node_i = 0
        self._interfaces = {}
        self._nodes = []

        self._service = service
        self._address = address
        self._path = path
        if not interface:
            self._interface = address
        else:
            self._interface = interface
        if changed_coroutines:
            self._changed_coroutines = changed_coroutines
        else:
            self._changed_coroutines = {}
        self._timeout_ms = timeout_ms
        self._camel_convert = camel_convert

    def __getattr__(self, name):
        if self._introspect_xml is None:
            raise RuntimeError("Must update proxy once before accessing.")
        return getattr(self._interfaces[self._interface], name)

    def __getitem__(self, interface):
        if self._introspect_xml is None:
            raise RuntimeError("Must update proxy once before accessing.")

        if interface not in self._interfaces:
            raise KeyError(f"Interface {interface} not in proxy")

        return self._interfaces[interface]

    async def __call__(self, node, changed_coroutines=None):
        if self._camel_convert:
            node = sdbus.snake_to_camel(node)

        if node not in self._nodes:
            raise AttributeError(f"Node {node} not in proxy")

        new = type(self)(
                self._service, self._address, f"{self._path}/{node}", None,
                changed_coroutines, self._timeout_ms, self._camel_convert
        )

        await new.update()
        return new

    async def __aenter__(self):
        return await self._interfaces[self._interface].__aenter__()

    async def __aexit__(
            self, exception_type, exception_value, exception_traceback
    ):
        return await self._interfaces[
                self._interface
        ].__aexit__(exception_type, exception_value, exception_traceback)

    def __aiter__(self):
        return self

    async def __anext__(self):
        i = self._node_i
        if i >= len(self._nodes):
            raise StopAsyncIteration
        proxy = await self(self._nodes[self._node_i])
        self._node_i += 1
        return proxy

    async def update(self, timeout_ms=None):
        """Use Introspection on remote server to update the proxy.
            **Must be run once before using the Proxy.**
        """
        timeout_ms = timeout_ms if timeout_ms is not None else self._timeout_ms

        self._introspect_xml = await call(
                self._service,
                self._address,
                self._path,
                'org.freedesktop.DBus.Introspectable',
                'Introspect',
                response_signature="s",
                timeout_ms=timeout_ms,
        )

        self._update_interfaces()
        for interface in self._interfaces.values():
            await interface.update_properties()

        return self

    def _update_interfaces(self):
        self._interfaces = {}

        root = etree.fromstring(self._introspect_xml)
        for e in root.iter('interface'):
            name = e.attrib['name']
            self._interfaces[name] = Interface(
                    self, self._service, self._address, self._path, e,
                    self._camel_convert, self._timeout_ms,
                    self._changed_coroutines.get(name, None)
            )

        for e in root.iter('node'):
            try:
                self._nodes.append(e.attrib['name'])
            except KeyError:
                pass



# Copyright: 2017, CCX Technologies
"""D-Bus Service"""

from . import sdbus


class Service:
    """Serves objects onto a D-Bus, runs within an asyncio loop.

    This is a class can be used to create a server that attaches the to
    D-Bus and either serves objects or provides and interface for a client
    proxy. To process requests it must be run within an asyncio loop.

    Args:
        name (str): name to be used on the D-Bus, ie. org.test
        loop: optional, asyncio loop to attach to, if not
            defined will use asyncio.get_event_loop()
        bus (str): optional, Bus to attach to either 'system', or 'session'
        replace_existing (bool): optional, if true and there is a
            D-Bus service with the same name, try to replace it
        allow_replacement (bool): optional, if true and another service
            tries to steal our name, let it
        name_queue (bool): optional, if our name is in use by another
            service put ourselves in the D-Bus name queue, we will get
            name once the existing service relinquishes it

    Raises:
        BusError: if an error occurs during initialization
    """
    def __init__(
            self,
            name=None,
            loop=None,
            bus='system',
            replace_existing=False,
            allow_replacement=False,
            name_queue=False
    ):
        self.sdbus = sdbus.Service(
                name, loop, bus, replace_existing, allow_replacement,
                name_queue
        )
        """Interface to sd-bus library"""

    def is_running(self):
        return self.sdbus.is_running()

    def get_loop(self):
        return self.sdbus.get_loop()

    def stop(self):
        self.sdbus.stop()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.stop()



# Copyright: 2017, CCX Technologies
"""D-Bus Signal"""

from .. import sdbus


class Signal:
    """Provides a method to emit a signal via D-Bus.

    This class is to be used as a Decorator for signals in
    an adbus.server.Object which will be exported via the D-Bus.

    Args:
        name (str): optional, signal name used in the D-Bus, if None the
            signal's label will be used
        deprecated (bool): optional, if true object is labelled
            as deprecated in the introspect XML data
        hidden (bool): optional, if true object won't be added
            to the introspect XML data
        camel_convert (bool): optional, D-Bus method and property
            names are typically defined in Camel Case, but Python
            methods and arguments are typically defined in Snake
            Case, if this is set the cases will be automatically
            converted between the two
    """
    def __init__(
            self,
            name=None,
            deprecated=False,
            hidden=False,
            camel_convert=True
    ):

        self.dbus_name = name
        self.deprecated = deprecated
        self.hidden = hidden
        self.camel_convert = camel_convert

    def __get__(self, instance, owner):
        raise RuntimeError("Can't read from a signal")

    def __set__(self, instance, value):
        signal = instance.__dict__[self.py_name]

        if isinstance(self.dbus_signature, list):
            signal.emit(*value)
        else:
            signal.emit(value)

    def __set_name__(self, owner, name):
        self.py_name = name

        try:
            signature = iter(owner.__annotations__[name])
        except TypeError:
            self.dbus_signature = (
                    sdbus.dbus_signature(owner.__annotations__[name]),
            )
        except KeyError:
            self.dbus_signature = sdbus.variant_signature()
        else:
            try:
                self.dbus_signature = [
                        sdbus.dbus_signature(s) for s in signature
                ]
            except (RuntimeError, TypeError):
                # this is a work around for an issue with the
                # typing library on Python version 3.7 creating an
                # unusable iterator from typing.List[int]
                self.dbus_signature = (
                        sdbus.dbus_signature(owner.__annotations__[name]),
                )

        if not self.dbus_name:
            self.dbus_name = name

        if self.camel_convert:
            self.dbus_name = sdbus.snake_to_camel(self.dbus_name)

    def vt(self, instance):
        signal = sdbus.Signal(
                self.dbus_name, self.dbus_signature, self.deprecated,
                self.hidden
        )
        instance.__dict__[self.py_name] = signal
        return signal



# Copyright: 2017, CCX Technologies
#cython: language_level=3

def snake_to_camel(snake):
    """Converts a snake_case string to CamelCase.

    Args:
        snake (str): Underscore separated string

    Returns:
        A string in CamelCase.
    """
    return "".join(x[:1].upper() + x[1:] for x in snake.split("_"))

first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')
def camel_to_snake(camel):
    """Converts CamelCase separated string to snake_case.

    Args:
        camel (str): CamelCase separated string

    Returns:
        A string in snake_case.
    """
    s1 = first_cap_re.sub(r'\1_\2', camel)
    return all_cap_re.sub(r'\1_\2', s1).lower()



# == Copyright: 2017, CCX Technologies

from asyncio import wait_for
from .. import sdbus
from .. import datatypes


async def get(service, address, path, interface, name, timeout_ms=30000):
    """Gets a D-Bus Property from another process.

    This is a co-routine, so must be await-ed from within a asyncio mainloop.

    Args:
        service (adbus.server.Service): service to connect to
        address (str): address (name) of the D-Bus Service to call
        path (str): path of method to call, ie. /com/awesome/Settings1
        interface (str): interface label to call, ie. com.awesome.settings
        name (str): name of the name to get, ie. TestProperty
        timeout_ms (int): maximum time to wait for a response in milli-seconds

    Returns:
        Value of the name.
    """

    call = sdbus.Call(
            service.sdbus, address.encode(), path.encode(),
            b"org.freedesktop.DBus.Properties", b"Get", (interface, name), b"v"
    )

    call.send(timeout_ms)
    await call.event.wait()

    if isinstance(call.response, Exception):
        raise call.response
    else:
        return call.response


async def get_all(service, address, path, interface, timeout_ms=30000):
    """Gets a All D-Bus Properties from another process.

    This is a co-routine, so must be await-ed from within a asyncio mainloop.

    Args:
        service (adbus.server.Service): service to connect to
        address (str): address (name) of the D-Bus Service to call
        path (str): path of method to call, ie. /com/awesome/Settings1
        interface (str): interface label to call, ie. com.awesome.settings
        timeout_ms (int): maximum time to wait for a response in milli-seconds

    Returns:
        A dictionary, keys are the name names and values are the name
        values.
    """

    call = sdbus.Call(
            service.sdbus, address.encode(), path.encode(),
            b"org.freedesktop.DBus.Properties", b"GetAll", (interface, ),
            b"a{sv}"
    )

    call.send(timeout_ms)
    await call.event.wait()

    if isinstance(call.response, Exception):
        raise call.response
    else:
        return call.response


async def set_(
        service, address, path, interface, name, value, timeout_ms=30000
):
    """Sets a D-Bus Property in another process.

    This is a co-routine, so must be await-ed from within a asyncio mainloop.

    Args:
        service (adbus.server.Service): service to connect to
        address (str): address (name) of the D-Bus Service to call
        path (str): path of method to call, ie. /com/awesome/Settings1
        interface (str): interface label to call, ie. com.awesome.settings
        name (str): name of the name to get, ie. TestProperty
        value (object): value to set the name to, must be compatible
            with the defined D-Bus Property type
        timeout_ms (int): maximum time to wait for a response in milli-seconds
    """

    call = sdbus.Call(
            service.sdbus, address.encode(), path.encode(),
            b"org.freedesktop.DBus.Properties", b"Set",
            (interface, name, datatypes.VariantWrapper(value))
    )

    call.send(timeout_ms)
    await call.event.wait()

    if isinstance(call.response, Exception):
        raise call.response



#!/usr/bin/python

from setuptools import setup
from setuptools import find_packages
from setuptools import Extension

try:
    from Cython.Build import cythonize
    USE_CYTHON = True
except ImportError:
    USE_CYTHON = False

__module__ = 'adbus'
__url__ = 'https://github.com/ccxtechnologies'

__version__ = None
exec(open(f'{__module__}/__version__.py').read())

ext = '.pyx' if USE_CYTHON else '.c'

module = [
        Extension(
                f"{__module__}.sdbus",
                sources=[f"{__module__}/sdbus" + ext],
                libraries=["systemd"],
                extra_compile_args=["-O3", "-std=c99"],  # match systemd
        )
]

if USE_CYTHON:
    module = cythonize(module)

setup(
        name=__module__,
        version=__version__,
        author='CCX Technologies',
        author_email='charles@ccxtechnologies.com',
        description='asyncio based dbus interface',
        long_description=open('README.rst', 'rt').read(),
        license='MIT',
        url=f'{__url__}/{__module__}',
        download_url=f'{__url__}/archive/v{__version__}.tar.gz',
        python_requires='>=3.7',
        packages=find_packages(exclude=["tests"]),
        setup_requires=[
                'setuptools>=18.0',  # Handles Cython extensions natively
                'cython>=0.25.2',
        ],
        ext_modules=module,
)



# == Copyright: 2017, CCX Technologies
#cython: language_level=3

cdef class Manager:
    cdef sdbus_h.sd_bus *bus
    cdef sdbus_h.sd_bus_slot *_slot
    cdef bytes path

    def __cinit__(self, service, path):
        self.path = path.encode()
        self.bus = (<Service>service).bus

        ret = sdbus_h.sd_bus_add_object_manager(self.bus, &self._slot, self.path)
        if ret < 0:
            raise SdbusError(f"Failed to add manager: {errorcode[-ret]}", -ret)

    def __dealloc__(self):
        self._slot = sdbus_h.sd_bus_slot_unref(self._slot)



# == Copyright: 2017, CCX Technologies
#cython: language_level=3

cdef int call_callback(sdbus_h.sd_bus_message *m, void *userdata,
        sdbus_h.sd_bus_error *err):
    cdef PyObject *call_ptr = <PyObject*>userdata
    cdef Call call = <Call>call_ptr
    cdef Message message = Message()
    cdef int errcode

    try:
        message.import_sd_bus_message(m)
        response = message.read(call.response_signature)
        if len(response) == 1:
            call.response = response[0]
        elif len(response) == 0:
            call.response = None
        else:
            call.response = response
        return 0

    except SdbusError as e:
        call.response = e
        errcode = <int>e.errno
        return -errcode

    finally:
        call.wake()
        Py_DECREF(call)

cdef class Call:
    cdef Message message
    cdef Service service
    cdef public object event
    cdef public object response
    cdef object wait_co
    cdef char *response_signature

    def __cinit__(self, Service service, address, path, interface, method,
            args=None, response_signature=b''):

        self.event = Event()
        self.service = service
        self.response = None
        self.message = Message()
        self.message.new_method_call(service, address, path, interface, method)
        self.response_signature = response_signature

        if args:
            for arg in args:
                signature = _dbus_signature(arg)
                self.message.append(signature, arg)

    def send(self, stdint.uint64_t timout_ms):
        cdef int ret
        self.event.clear()

        Py_INCREF(self)
        ret = sdbus_h.sd_bus_call_async(self.service.bus, NULL,
                self.message.message, call_callback, <void *>self,
                timout_ms*1000)
        if ret < 0:
            Py_DECREF(self)
            self.event.set()
            raise SdbusError(f"Failed to send call: {errorcode[-ret]}", -ret)

    cdef wake(self):
        self.event.set()



# Copyright: 2017-2021, CCX Technologies
#cython: language_level=3

cdef class Service:
    cdef sdbus_h.sd_bus *bus
    cdef bytes name
    cdef bytes unique_name
    cdef bool connected
    cdef bool process_loop
    cdef stdint.uint64_t flags
    cdef object loop
    cdef object startup_task

    def __cinit__(self, name=None, loop=None, bus='system',
            replace_existing=False, allow_replacement=False, queue=False):
        cdef const char *unique

        if name:
            self.name = name.encode()
        else:
            self.name = b''
        self.flags = 0
        self.connected = False
        self.loop = loop

        if bus == 'system':
            if sdbus_h.sd_bus_open_system(&self.bus) < 0:
                raise BusError("Failed to connect to Bus")

        elif bus == 'session':
            if sdbus_h.sd_bus_open_user(&self.bus) < 0:
                raise BusError("Failed to connect to Bus")

        else:
            raise BusError(f"Invalid bus {bus}, expecting system of session")

        if replace_existing:
            self.flags |= sdbus_h.SD_BUS_NAME_REPLACE_EXISTING

        if allow_replacement:
            self.flags |= sdbus_h.SD_BUS_NAME_ALLOW_REPLACEMENT

        if queue:
            self.flags |= sdbus_h.SD_BUS_NAME_QUEUE

        if self.name:
            if sdbus_h.sd_bus_request_name(self.bus, self.name, self.flags) < 0:
                raise BusError(f"Failed to acquire name {self.name.decode('utf-8')}")

        if sdbus_h.sd_bus_get_unique_name(self.bus, &unique) < 0:
            raise BusError("Failed to read unique name")
        self.unique_name = unique

        bus_fd = sdbus_h.sd_bus_get_fd(self.bus)
        if bus_fd <= 0:
            raise BusError("Failed to read sd-bus file descriptor")

        if not loop:
            loop = get_event_loop()

        self.process_loop = True
        self.startup_task = loop.create_task(self.startup_process())
        loop.add_reader(bus_fd, self.process)

        self.loop = loop

    def stop(self):
        self.startup_task.cancel()
        bus_fd = sdbus_h.sd_bus_get_fd(self.bus)
        if bus_fd > 0:
            self.loop.remove_reader(bus_fd)
        self.process_loop = False

        self.bus = sdbus_h.sd_bus_flush_close_unref(self.bus)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.stop()

    def is_running(self):
        """Service is running."""
        return self.loop.is_running()

    def get_loop(self):
        """Get the service's asyncio loop."""
        return self.loop

    async def startup_process(self):
        self.process()

    def process(self):
        """Processes all available transactions from the D-Bus.

        Raises:
            sdbus.BusError: if an error occurs interfacing with the D-Bus
            Exception: re-raised if raised by any callbacks, or property setters
        """

        if self.connected:
            raise BusError(f"Already processing")

        self.connected = True
        try:
            while self.process_loop:
                r = sdbus_h.sd_bus_process(self.bus, NULL)

                if r < 0:
                    raise BusError(f"D-Bus Process Error: {errorcode[-r]}")

                if r == 0:
                    break

        finally:
            self.connected = False



# == Copyright: 2017, CCX Technologies

import inspect

from .. import sdbus


class Listen:
    """Listens for a D-Bus Signal from another process.

    Args:
        service (adbus.server.Service): service to connect to
        address (str): address (name) of the D-Bus Service to call
        path (str): path of method to call, ie. /com/awesome/Settings1
        interface (str): interface label to call, ie. com.awesome.settings
        signal (str): name of the signal to listen to, ie. TestSignal
        coroutine (coroutine): coroutine to schedule when signal is received
        args (list or tuple): optional, list of argument values to match,
            the argument must be a string, useful for listening to property
            changes
        signature (str): optional, signature of the signal, if False the
            types of the coroutine arguments will be used to create the
            signature and the coroutine will be called with one argument
            per signal argument, if defined the coroutine will be called
            with a list of arguments, if None the coroutine will be called
            with a list of types determined at run-time
    """
    def __init__(
            self,
            service,
            address,
            path,
            interface,
            signal,
            coroutine,
            args=(),
            signature=False,
    ):

        if signature is False:
            self.signature = ''
            sig = inspect.signature(coroutine)
            for param in sig.parameters.values():
                if param.annotation != inspect.Parameter.empty:
                    self.signature += sdbus.dbus_signature(param.annotation)
                else:
                    self.signature += sdbus.variant_signature()

        elif signature is None:
            self.signature = 'ANY'

        else:
            self.signature = signature

        try:
            self.sdbus = sdbus.Listen(
                    service.sdbus, address, path, interface, signal, coroutine,
                    args, self.signature.encode(), signature is False
            )
        except sdbus.SdbusError:
            # sometimes we'll get a EINTR (like one in a million trys), we want
            # to try again if we do
            self.sdbus = sdbus.Listen(
                    service.sdbus, address, path, interface, signal, coroutine,
                    args, self.signature.encode(), signature is False
            )
